#include <math.h>
#include <stdlib.h>                      // System includes
#include <stdio.h>
#include <sys/time.h>
#include <unistd.h>
#include "marcel.h"
//sin(xy) - 1/2 = 0
//y^2 - 6x - 2  = 0


/* *********************************************
 * AD: tentative d'amlioration de la stabilit
 */
#define alta_malloc TBX_MALLOC
#define alta_free   TBX_FREE
#define alta_printf DISP

#define  h ((double)180)

#define  hh ((double)h/1e3)

#define  max_iter 10000
#define ERROR_THRE 1


#define  overlapx  10
#define  overlapz  10
#define TAUX_PERTE 50  // definition du taux de perte authorise
#define T_CONV     1  // 0 pour centralis 1 pour decentralis
#define T_BARR     1  // 0 barrire normale 1 autre (deffectueuse)

#define SQR(A)    ((A) * (A))

#define NBCONV 3

#define NUM_SPECIES  2             /* number of species         */
#define KH           4.0e-6        /* horizontal diffusivity Kh */
#define VEL          0.001         /* advection velocity V      */
#define KV0          1.0e-8        /* coefficient in Kv(z)      */
#define Q1           1.63e-16      /* coefficients q1, q2, c3   */ 
#define Q2           4.66e-16
#define C3           3.7e16
#define A3           22.62         /* coefficient in expression for q3(t) */
#define A4           7.601         /* coefficient in expression for q4(t) */
#define C1_SCALE     1.0e6         /* coefficients in initial profiles    */
#define C2_SCALE     1.0e12

#define T0           0.0           /* initial time */
#define NOUT         3            /* number of output times */
#define TWOHR        720.0       /* number of seconds in two hours  */
#define HALFDAY      4.32e4       /* number of seconds in a half day */
#define PI       3.1415926535898   /* pi */ 

#define XMIN          0.0          /* grid boundaries in x  */
#define XMAX         40.0           
#define ZMIN         20.0          /* grid boundaries in z  */
#define ZMAX         60.0
#define XMID         20.0          /* grid midpoints in x,z */          
#define ZMID         40.0

int NPEX;
int NPEZ;


int MXSUB;
int MZSUB;
int MX;
int MZ;

//#define MXSUB           540           /* MX = number of x mesh points */
//#define MZSUB           60           /* MZ = number of z mesh points */
//#define NSMXSUB         NUM_SPECIES*MXSUB

//#define MX           (NPEX*MXSUB)   /* MX = number of x mesh points */
//#define MZ           (NPEZ*MZSUB)   /* MY = number of y mesh points */



/* CVodeMalloc Constants */

#define RTOL    1.0e-5            /* scalar relative tolerance */
#define FLOOR   100.0             /* value of C1 or C2 at which tolerances */
                                  /* change from relative to absolute      */
#define ATOL    (RTOL*FLOOR)      /* scalar absolute tolerance */
int  NEQ=0;      /* NEQ = number of equations */

#define IJKth(vdata,i,j,k) (vdata[i-1 + (j)*NUM_SPECIES + (k)*(data.cex*NUM_SPECIES)])


#define COMPLETION 1000
#define CONVERGENCE 1001
#define DATA 1002

typedef struct {
  double q4, om, dx, dz, hdco, haco, vdco;
  long myId, isubx, isubz;
  int csx,cex,csz,cez;   //compute start and end for x and z
  int bx,bz;         //base for x and y, required for the overlapping
  int dsx,dex,dsz,dez; //display start and end for x and z
  int offx,offz; //offset of x and z for the start to computation
} UserData;

double *val;
int *colptr;
int *rowind;
int nb_non_zero=0;
double tstart;
UserData data;
double *vec_y,*vec_yy,*vec_b,*vec_r,*tmp,*b,*r,*tmp_y,*tmp_yy,*new_y,*old_y,*y;

int nprocs,myId;
static marcel_mutex_t  mutexst = MARCEL_MUTEX_INITIALIZER;
static marcel_mutex_t  mutexsb = MARCEL_MUTEX_INITIALIZER;
static marcel_mutex_t  mutexsl = MARCEL_MUTEX_INITIALIZER;
static marcel_mutex_t  mutexsr = MARCEL_MUTEX_INITIALIZER;
double *tab_errors;
int stop_iteration=0;
int nbIterCV=0;
double error,maxerror;
int convergence;
marcel_t thsb, thst,thsl, thsr, thrl, thrr,thrb, thrt,thrc;
int stop=0;


/* Passage Marcel*/

typedef struct {
  int *tab_sync;
  marcel_mutex_t sync_mutex;
}  barrier_t;  // temporaire pm2.h


typedef enum {CONV, STOP, SYNCHRO,SYNCHRO_D, CANCEL_PARTIAL_CV, PARTIAL_CV, GLOBAL_CV, FIN}t_ordre;
typedef enum{TOP, BOTTOM, RIGHT, LEFT, ME}t_position;

barrier_t barrier; 
p_mad_channel_t * channel; //tableau des sous-canaux
static marcel_t * thread_ecoute;
static int nb_lost_msgs  =   0;
int reception;
int quit = 0;
int barrierC=0;    //controle numero de barriere


#define THRESHOLD_SLCV NBCONV
int nbNeighbors;
int nbNeigNotLCV;
int  nbIterPreLocalCV;
int * prevIterNumS;
int * prevIterNumC;
int * stop_calcul;
int * neighbors;
int destinataire;
_Bool  preLocalCV;
_Bool  sLocalCV;
_Bool  globalCV;


/* Fin Marcel */

#define maxloc(a,b) (a)>(b)?(a):(b)
#define minloc(a,b) (a)<(b)?(a):(b)

int gmres_solve(int size, int nb_non_zero, double *val, int *colptr, int *rowind, double* b, double *r);


double TimeStart()
{
  struct timeval tstart;
  gettimeofday(&tstart,0);
  return( (double) (tstart.tv_sec + tstart.tv_usec*1e-6) );
}

double TimeStop(double t)
{
  struct timeval tend;

  gettimeofday(&tend,0);
  t = (double) (tend.tv_sec + tend.tv_usec*1e-6) - t;
  return (t);
}



void InitUserData()
{
  int isubx,isubz;
  data.om = PI/HALFDAY;
  data.dx = (XMAX-XMIN)/(MX-1);
  data.dz = (ZMAX-ZMIN)/(MZ-1);
  data.hdco = KH/SQR(data.dx);
  data.haco = VEL/(2.0*data.dx);
  data.vdco = (1.0/SQR(data.dz))*KV0;

  data.myId = myId;
  /* isubx and isuby are the PE grid indices corresponding to my_pe */
  isubz = myId/NPEX;
  isubx = myId - isubz*NPEX;
  data.isubx = isubx;
  data.isubz = isubz;
  data.csx=1;
  data.cex=MXSUB+2;
  data.csz=1;
  data.cez=MZSUB+2;
  data.offx=isubx*MXSUB;
  data.offz=isubz*MZSUB;
  data.bx=0;
  data.bz=0;
  data.dsx=0;
  data.dex=MXSUB;
  data.dsz=0;
  data.dez=MZSUB;

  if(NPEZ>1) {
    if(data.isubz==0) {
      data.cez+=overlapz;
    }else{
      data.dsz+=overlapz;
      data.dez+=overlapz;
      if(data.isubz==NPEZ-1) {
	data.cez+=overlapz;
	data.bz-=overlapz;
      }else{
	data.cez+=2*overlapz;
	data.bz-=overlapz;
      }
    }
  }

  if(NPEX>1) {
    if(data.isubx==0) {
      data.cex+=overlapx;
    }else{
      data.dsx+=overlapx;
      data.dex+=overlapx;
      if(data.isubx==NPEX-1) {
	data.cex+=overlapx;
	data.bx-=overlapx;
      }else{
	data.cex+=2*overlapx;
	data.bx-=overlapx;
      }
    }
  }

  alta_printf("%ld %ld\n",data.isubx,data.isubz);

  NEQ=(NUM_SPECIES*(data.cex)*(data.cez));

}



static void SetInitialProfiles(double* y)
{
  int jx, jz, lx,lz;
  double x, z, cx, cz;
  /* Set pointer to data array in vector y. */


  for (lz=data.csz; lz < data.cez-1; lz++) {
    jz= (lz-1+data.bz)+data.offz;
    z = ZMIN + jz*data.dz;
    cz = SQR(0.1*(z - ZMID));
    cz = 1.0 - cz + 0.5*SQR(cz);
    for (lx=data.csx; lx < data.cex-1; lx++) {
      jx = (lx-1+data.bx) + data.offx;
      x = XMIN + jx*data.dx;
      cx = SQR(0.1*(x - XMID));
      cx = 1.0 - cx + 0.5*SQR(cx);
      IJKth(y,1,lx,lz) = C1_SCALE*cx*cz; 
      IJKth(y,2,lx,lz) = C2_SCALE*cx*cz;
    }
  }
}

int is_on_grid(int index) {
  int specie;
  int jx,jz;
  specie=index%NUM_SPECIES;
  index/=NUM_SPECIES;
  jx=index%(data.cex);
  jz=index/(data.cex);

  return jx>0 && jx< data.cex-1&& jz>0 && jz<data.cez-1;
}

void ff(double*dydt ,double* y, double *old_y,double t)
{
  double q3, c1, c2, c1dn, c2dn, c1up, c2up, c1lt, c2lt;
  double c1rt, c2rt, czdn, czup, hord1, hord2, horad1, horad2;
  double qq1, qq2, qq3, qq4, rkin1, rkin2, s, vertd1, vertd2, zdn, zup;
  double q4coef, delz, verdco, hordco, horaco;
  int jx, jz, lx, lz, idn, iup, ileft, iright;

  /* Set diurnal rate coefficients. */

  s = sin(data.om*t);
  if (s > 0.0) {
    q3 = exp(-A3/s);
    data.q4 = exp(-A4/s);
  } else {
    q3 = 0.0;
    data.q4 = 0.0;
  }

  /* Make local copies of problem variables, for efficiency. */

  q4coef = data.q4;
  delz = data.dz;
  verdco = data.vdco;
  hordco  = data.hdco;
  horaco  = data.haco;

  /* Loop over all grid points. */

  for (lz=data.csz; lz < data.cez; lz++) {    //attention +1 pour les extrmits

    jz = (lz-1+data.bz) + data.offz;          //on retranche le 1
    zdn = ZMIN + (jz - .5)*delz;
    zup = zdn + delz;
    czdn = verdco*exp(0.2*zdn);
    czup = verdco*exp(0.2*zup);
    idn = (jz == 0) ? 1 : -1;    //test sur l'indice global
    iup = (jz == MZ-1) ? -1 : 1; //idem
    for (lx=data.csx; lx < data.cex; lx++) {  //attention +1 pour les extrmits
      jx = (lx-1+data.bx) + data.offx;

      c1 = IJKth(y,1,lx,lz); 
      c2 = IJKth(y,2,lx,lz);
      qq1 = Q1*c1*C3;
      qq2 = Q2*c1*c2;
      qq3 = q3*C3;
      qq4 = q4coef*c2;
      rkin1 = -qq1 - qq2 + 2.0*qq3 + qq4;
      rkin2 = qq1 - qq2 - qq4;
      /* Set vertical diffusion terms. */

      c1dn = IJKth(y,1,lx,lz+idn);
      c2dn = IJKth(y,2,lx,lz+idn);
      c1up = IJKth(y,1,lx,lz+iup);
      c2up = IJKth(y,2,lx,lz+iup);
      vertd1 = czup*(c1up - c1) - czdn*(c1 - c1dn);
      vertd2 = czup*(c2up - c2) - czdn*(c2 - c2dn);

      /* Set horizontal diffusion and advection terms. */

      ileft = (jx == 0) ? 1 : -1;
      iright =(jx == MX-1) ? -1 : 1;
      c1lt = IJKth(y,1,lx+ileft,lz); 
      c2lt = IJKth(y,2,lx+ileft,lz);
      c1rt = IJKth(y,1,lx+iright,lz);
      c2rt = IJKth(y,2,lx+iright,lz);
      hord1 = hordco*(c1rt - 2.0*c1 + c1lt);
      hord2 = hordco*(c2rt - 2.0*c2 + c2lt);
      horad1 = horaco*(c1rt - c1lt);
      horad2 = horaco*(c2rt - c2lt);


      /* Load all terms into ydot. */
      //			cout<<vertd1 + hord1 + horad1 + rkin1<<endl;
      IJKth(dydt, 1, lx, lz) = -IJKth(old_y, 1, lx, lz)-h*(vertd1 + hord1 + horad1 + rkin1)+IJKth(y, 1, lx, lz); 
      IJKth(dydt, 2, lx, lz) = -IJKth(old_y, 2, lx, lz)-h*(vertd2 + hord2 + horad2 + rkin2)+IJKth(y, 2, lx, lz);

      //if(myId==0)
      //	cout<<jx<<" "<<jz<<" "<<IJKth(dydt, 1, jx, jz)<<" "<<IJKth(dydt, 2, jx, jz)<<endl;

    }
  }
}


void ff_part(double*dydt ,double* y, double *old_y, int index, double t)
{
  double q3, c1, c2, c1dn, c2dn, c1up, c2up, c1lt, c2lt;
  double c1rt, c2rt, czdn, czup, hord1, hord2, horad1, horad2;
  double qq1, qq2, qq3, qq4, rkin1, rkin2, s, vertd1, vertd2, zdn, zup;
  double q4coef, delz, verdco, hordco, horaco;
  int jx, jz,lx,lz, idn, iup, ileft, iright;
  int specie;
  int ind=index;
  specie=ind%NUM_SPECIES;
  ind/=NUM_SPECIES;
  lx=ind%(data.cex);
  lz=ind/(data.cex);
  /* Set diurnal rate coefficients. */


  s = sin(data.om*t);
  if (s > 0.0) {
    q3 = exp(-A3/s);
    data.q4 = exp(-A4/s);
  } else {
    q3 = 0.0;
    data.q4 = 0.0;
  }

  /* Make local copies of problem variables, for efficiency. */

  q4coef = data.q4;
  delz = data.dz;
  verdco = data.vdco;
  hordco  = data.hdco;
  horaco  = data.haco;

  /* Loop over all grid points. */

  //for (jz=0; jz < MZ; jz++) 
  {

    /* Set vertical diffusion coefficients at jz +- 1/2 */
    jz = (lz-1+data.bz) + data.offz; 
    zdn = ZMIN + (jz - .5)*delz;
    zup = zdn + delz;
    czdn = verdco*exp(0.2*zdn);
    czup = verdco*exp(0.2*zup);
    idn = (jz == 0) ? 1 : -1;
    iup = (jz == MZ-1) ? -1 : 1;
    //for (jx=0; jx < MX; jx++) 
    {
      jx = (lx-1+data.bx) + data.offx; 
      /* Extract c1 and c2, and set kinetic rate terms. */
      //if(myId==0)
      //	cout<<"arg "<<specie<<" "<<lx<<" "<<lz<<endl;
	

      c1 = IJKth(y,1,lx,lz); 
      c2 = IJKth(y,2,lx,lz);
      //			cout<<"c "<<c1<<" "<<c2<<endl;
      qq1 = Q1*c1*C3;
      qq2 = Q2*c1*c2;
      qq3 = q3*C3;
      qq4 = q4coef*c2;
      rkin1 = -qq1 - qq2 + 2.0*qq3 + qq4;
      rkin2 = qq1 - qq2 - qq4;
      //cout<<"rk "<<rkin1<<" "<<rkin2<<endl;
      /* Set vertical diffusion terms. */

      c1dn = IJKth(y,1,lx,lz+idn);
      c2dn = IJKth(y,2,lx,lz+idn);
      c1up = IJKth(y,1,lx,lz+iup);
      c2up = IJKth(y,2,lx,lz+iup);
      vertd1 = czup*(c1up - c1) - czdn*(c1 - c1dn);
      vertd2 = czup*(c2up - c2) - czdn*(c2 - c2dn);

      /* Set horizontal diffusion and advection terms. */

      ileft = (jx == 0) ? 1 : -1;
      iright =(jx == MX-1) ? -1 : 1;
      c1lt = IJKth(y,1,lx+ileft,lz); 
      c2lt = IJKth(y,2,lx+ileft,lz);
      c1rt = IJKth(y,1,lx+iright,lz);
      c2rt = IJKth(y,2,lx+iright,lz);
      hord1 = hordco*(c1rt - 2.0*c1 + c1lt);
      hord2 = hordco*(c2rt - 2.0*c2 + c2lt);
      horad1 = horaco*(c1rt - c1lt);
      horad2 = horaco*(c2rt - c2lt);

      /* Load all terms into ydot. */
      //			cout<<vertd1 + hord1 + horad1 + rkin1<<endl;
      IJKth(dydt, 1, lx, lz) = -IJKth(old_y, 1, lx, lz)-h*(vertd1 + hord1 + horad1 + rkin1)+IJKth(y, 1, lx, lz); 
      IJKth(dydt, 2, lx, lz) = -IJKth(old_y, 2, lx, lz)-h*(vertd2 + hord2 + horad2 + rkin2)+IJKth(y, 2, lx, lz);

      //if(myId==0)
      //	cout<<jx<<" "<<jz<<" "<<index<<" "<<IJKth(dydt, 1, lx, lz)<<" "<<IJKth(dydt, 2, lx, lz)<<" "<<2*lx+(MXSUB+2)*2*lz<<endl;
    }
  }
}

inline void deriv(int j) {

  int line=(j/NUM_SPECIES)/(data.cex);


  double v=(tmp[j]-vec_b[j])/hh;
  if(v!=0) {
    int jloc=j-NUM_SPECIES*(data.cex-1)-(line*2)*NUM_SPECIES;  //on retranche une ligne - 1 cot et on retranche les autres cots pour les autres lignes                      //essai (MXSUB+1)
    if(jloc<(data.cex-2)*(data.cez-2)*NUM_SPECIES && jloc>= 0) {
      val[nb_non_zero]=v;
      rowind[nb_non_zero++]=jloc;  
    }
  }
}


void jacobian(int index, double* tmp, double* vec_yy, double* old_y, double t) {
  int specie;
  int lx,lz,jx,jz;
  int idn,iup;
  int ileft,iright;

  //index=32;
  int ind=index;
  specie=ind%NUM_SPECIES;
  ind/=NUM_SPECIES;
  lx=ind%(data.cex);
  lz=ind/(data.cex);

  jx = lx + data.offx;
  jz = lz + data.offz;
  idn = (jz-1 == 0) ? 1 : -1;
  iup = (jz-1 == MZ-1) ? -1 : 1;

  /*if(myId==0)
    cout<<"ind "<<index<<" "<<lx<<" "<<lz<<" "<<jx<<" "<<jz<<endl;
    if(myId==0)
    cout<<"before "<<tmp[index]<<endl;*/
  ff_part(tmp, vec_yy, old_y, index, t);
  //if(myId==0)
  //	cout<<"after "<<tmp[index]<<endl;
  deriv(index);
	
  if(specie==0)
    deriv(index+1);
  else
    deriv(index-1);

  index=specie + (lx)*NUM_SPECIES + (lz+idn)*(NUM_SPECIES*(data.cex));

  if(is_on_grid(index)) {
    ff_part(tmp, vec_yy, old_y, index, t);
    deriv(index);
    if(specie==0)
      deriv(index+1);
    else
      deriv(index-1);
  }

  if(iup!=idn) {
    index=specie + (lx)*NUM_SPECIES + (lz+iup)*(NUM_SPECIES*(data.cex));
    if(is_on_grid(index)) {
      ff_part(tmp, vec_yy, old_y, index, t);
      deriv(index);
      if(specie==0)
	deriv(index+1);
      else
	deriv(index-1);
    }
  }


  ileft = (jx-1 == 0) ? 1 : -1;
  iright =(jx-1 == MX-1) ? -1 : 1;
  index=specie + (lx+ileft)*NUM_SPECIES + (lz)*(NUM_SPECIES*(data.cex));
  if(is_on_grid(index)) {
    ff_part(tmp, vec_yy, old_y, index, t);
    deriv(index);
    if(specie==0)
      deriv(index+1);
    else
      deriv(index-1);
  }
	
  if(ileft!=iright) {
    index=specie + (lx+iright)*NUM_SPECIES + (lz)*(NUM_SPECIES*(data.cex));
    if(is_on_grid(index)) {
      ff_part(tmp, vec_yy, old_y, index, t);
      deriv(index);
      if(specie==0)
	deriv(index+1);
      else
	deriv(index-1);
    }
  }

}


/* Fonction communication */

#define barrier_lock(bar) marcel_mutex_lock(&bar->sync_mutex)
#define barrier_unlock(bar) marcel_mutex_unlock(&bar->sync_mutex)

static void send_fin()
{
  p_mad_connection_t  out    = NULL;
  int i=   (myId + 1)%(NPEX*NPEZ);
  t_ordre ordre = FIN;
  out  = mad_begin_packing(channel[0], i);
  mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_end_packing(out);
}

static void send_stop( int a)
{
  p_mad_connection_t  out    = NULL;
  int i=a;
  int st = 1;
  t_ordre ordre = STOP;
  out  = mad_begin_packing(channel[0], i);
  mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &st, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_end_packing(out);
}

static void send_c()
{
  p_mad_connection_t  out    = NULL;
  t_ordre ordre = CONV;
  out  = mad_begin_packing(channel[0], 0);
  mad_pack(out, &ordre,  sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &myId, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &error, sizeof(double), mad_send_CHEAPER, mad_receive_CHEAPER);
  mad_end_packing(out);
}


void send_bd(int id){
  p_mad_connection_t  out    = NULL;
  t_ordre ordre = SYNCHRO;
  
  out =  mad_begin_packing(channel[0], id);
  mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &myId, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_end_packing(out);
}


void send_bdd(int id, int fsender){
  p_mad_connection_t  out    = NULL;
  t_ordre ordre = SYNCHRO_D;
  
  out =  mad_begin_packing(channel[0], id);
  mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &myId, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_pack(out, &fsender, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_end_packing(out);
}


void send_cancelPartialCV(int iteration)
{
  p_mad_connection_t  out    = NULL;
  t_ordre ordre = CANCEL_PARTIAL_CV;
  
  if (destinataire != -1){
    out  = mad_begin_packing(channel[0], destinataire);
    mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
    mad_pack(out, &myId,sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
    mad_pack(out, &iteration,sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
    mad_end_packing(out);
    destinataire = -1;
  }
}



void send_PartialCV(int iteration)
{
  p_mad_connection_t  out    = NULL;
  int i;
  t_ordre ordre = PARTIAL_CV;
  
  for (i = 0; i<nbNeighbors && destinataire == -1; i++){
    if ((prevIterNumS[i] == -1) || (prevIterNumS[i]< prevIterNumC[i])){
      destinataire = neighbors[i];
      out  = mad_begin_packing(channel[0], destinataire); 
      mad_pack(out, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
      mad_pack(out, &myId,sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
      mad_pack(out, &iteration,sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
      mad_end_packing(out);
    }
  } 
}


void send_GlobalCV()
{
  int i;
  for (i = 0; i<nbNeighbors; i++){
    if (stop_calcul[i] == 0){
      send_stop(neighbors[i]);
    }
  }
}




static void rec_ordre(void)
{
  p_mad_connection_t  in     = NULL;
  t_ordre ordre;
  int index;
  int i,flag=1;
  int sender;
  int fsender;
  barrier_t *bar=&barrier;

  int iteration;
  int trouver;
  int indice;


  marcel_detach(marcel_self());
  while (quit == 0)
    {
      in = mad_begin_unpacking(channel[0]);
      mad_unpack(in, &ordre, sizeof(t_ordre), mad_send_CHEAPER, mad_receive_EXPRESS);
      switch (ordre){
       
      case CONV:
       
	mad_unpack(in, &index, sizeof(int), mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_unpack(in, &tab_errors[index], sizeof(double), mad_send_CHEAPER,mad_receive_CHEAPER);
	mad_end_unpacking(in);
	if(nbIterCV>=NBCONV) 	{
	  flag=1;
	  for(i=0;i<nprocs && flag;++i){
	    flag&=(tab_errors[i]<ERROR_THRE);
	  }
	  if(flag){
	    stop_iteration=1;
	  }
	  if(stop_iteration==1 && myId==0){
	    stop_iteration=2;
	    maxerror=error;
	    for(i=1;i<nprocs;++i){
	      if(maxerror<tab_errors[i])
		maxerror=tab_errors[i];
	    }
	    for(i=1;i<nprocs;++i){
	      send_stop(i);
	    }
	  }
	}
	  
	break;

      case STOP:
	
	mad_unpack(in, &stop_iteration, sizeof(int), mad_send_CHEAPER,mad_receive_EXPRESS);
	mad_end_unpacking(in);
	break;

      case SYNCHRO:
	mad_unpack(in, &sender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_end_unpacking(in);
 
	barrier_lock(bar);
	bar->tab_sync[sender]++;
	barrier_unlock(bar);

	break;

      case SYNCHRO_D:
	mad_unpack(in, &sender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_unpack(in, &fsender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_end_unpacking(in);
	
	barrier_lock(bar);
	bar->tab_sync[fsender]++;
	barrier_unlock(bar);
	
	if(myId > 0 && myId - 1 != sender ) { // left
	  //	  DISP("envoie a gauche %d ", fsender);
	  send_bdd(myId - 1, fsender);
	}
	if(myId < (NPEX* NPEZ)-1 && myId + 1 != sender ) { // right
	  //	  DISP("envoie a droite %d", fsender);
	  send_bdd(myId + 1, fsender);
	}
	
	break;


      case CANCEL_PARTIAL_CV:
	mad_unpack(in, &sender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_unpack(in, &iteration, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_end_unpacking(in);	

	trouver = 0;
	for (i=0; i<=nbNeighbors && trouver == 0; i++){
	  if (neighbors[i] ==  sender){
	    sender = i;
	    trouver = 1;
	  }
	}

	if (prevIterNumC[sender] < prevIterNumS[sender]&& prevIterNumS[sender] < iteration){
	  nbNeigNotLCV ++;
	  globalCV = FALSE;
	}
	if  (prevIterNumC[sender] < iteration ){
	  prevIterNumC[sender] = iteration;
	}
	break;

      case PARTIAL_CV:
	mad_unpack(in, &sender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_unpack(in, &iteration, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_end_unpacking(in);
	
	trouver = 0;
	for (i=0; i<=nbNeighbors && trouver == 0; i++){
	  if (neighbors[i] ==  sender){
	    sender = i;
	    trouver = 1;
	  }
	}

	if (prevIterNumS[sender] <= prevIterNumC[sender] && prevIterNumC[sender] <= iteration){
	  nbNeigNotLCV --;
	}
	if (prevIterNumS[sender] < iteration){
	  prevIterNumS[sender]=iteration;
	}
	break;

      case GLOBAL_CV:
	mad_unpack(in, &sender, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_unpack(in, &iteration, sizeof(int),mad_send_CHEAPER, mad_receive_EXPRESS);
	mad_end_unpacking(in);

	for (i=0; i<nbNeighbors; i++)
	  if (neighbors[i] == sender) indice = i;
	
	if (prevIterNumS[indice] <= prevIterNumC[indice] && prevIterNumC[indice] < iteration){
	  nbNeigNotLCV --;
		}
	if (prevIterNumS[indice] < iteration){
	  prevIterNumS[indice]=iteration;
	}
	break;

	
      case FIN:
	DISP("j'ai recu fin\n");
	break;
	
      default:
	break;
      }
    }
  alta_printf("thread ecoute ordre ferm\n");
  marcel_yield(); 
}

void send_t(void* dum)
{
  t_position place = TOP;
  p_mad_connection_t  out    = NULL;
  int offs=NUM_SPECIES*((data.cex)*(2*overlapz+1)+1);
  int size=(data.cex-2)*NUM_SPECIES;
  int i = 0;

  out = mad_begin_packing(channel[1], myId-NPEX);
  mad_pack(out, &place,sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);
  for (i=0; i<size;i++){
    p_mad_buffer_slice_parameter_t param = NULL;
    param = mad_alloc_slice_parameter();
    param->opcode = mad_op_optional_block;
    param->value  = TAUX_PERTE; /* prabibilit de perte */
    param->offset = 0; /* Pour l'instant inutile */
    param->length = 0; /* Pour l'instant inutile */
    mad_pack_ext(out,
		 &vec_y[offs+i], sizeof(double),
		 mad_send_CHEAPER, mad_receive_CHEAPER,
		 param, NULL); /* donnees */
  }
  mad_end_packing(out);
  marcel_mutex_unlock(&mutexst);
}

void send_b(void* dum)
{
  t_position place = BOTTOM;
  p_mad_connection_t  out    = NULL;
  int offs=NEQ-NUM_SPECIES*(data.cex-1+data.cex+2*overlapz*(data.cex));
  int size=(data.cex-2)*NUM_SPECIES;
  int i = 0;
  
  out = mad_begin_packing(channel[1], myId+NPEX);
  mad_pack(out, &place,sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);
  for (i=0; i<size;i++){
    p_mad_buffer_slice_parameter_t param = NULL;
    param = mad_alloc_slice_parameter();
    param->opcode = mad_op_optional_block;
    param->value  = TAUX_PERTE; /* prabibilit de perte */
    param->offset = 0; /* Pour l'instant inutile */
    param->length = 0; /* Pour l'instant inutile */
    mad_pack_ext(out,
		 &vec_y[offs+i], sizeof(double),
		 mad_send_CHEAPER, mad_receive_CHEAPER,
		 param, NULL); /* donnees */
  }
  mad_end_packing(out);
  marcel_mutex_unlock(&mutexsb);
}


void send_l(void* dum)
{
  t_position place = LEFT;
  p_mad_connection_t  out    = NULL;
  int offs=NUM_SPECIES*(data.cex+overlapx*2+1);
  int size=(data.cez-2)*NUM_SPECIES;
  double bufs[data.cez*NUM_SPECIES];
  int i,j,last;

  last=0;
  for(i=0;i<data.cez-2;i++) {
    for(j=0;j<NUM_SPECIES;j++)
      bufs[last+j]=vec_y[offs+j];
    last+=NUM_SPECIES;
    offs+=NUM_SPECIES*(data.cex);
  }
  
  out = mad_begin_packing(channel[1], myId-1);
  mad_pack(out, &place,sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);
  for (i=0; i<size; i++){
    p_mad_buffer_slice_parameter_t param = NULL;
    param = mad_alloc_slice_parameter();
    param->opcode = mad_op_optional_block;
    param->value  = TAUX_PERTE; /* prabibilit de perte */
    param->offset = 0; /* Pour l'instant inutile */
    param->length = 0; /* Pour l'instant inutile */
    mad_pack_ext(out,
		&bufs[i], sizeof(double),
		 mad_send_CHEAPER, mad_receive_CHEAPER,
		 param, NULL); /* donnees */
   }
  mad_end_packing(out);
  marcel_mutex_unlock(&mutexsl);
}

void send_r(void* dum)
{
  t_position place = RIGHT;
  p_mad_connection_t  out    = NULL;
  int offs=NUM_SPECIES*(2*data.cex-1-2*overlapx);
  int size=(data.cez-2)*NUM_SPECIES;
  double bufs[data.cez*NUM_SPECIES];

  int i,j,last;
  
  last=0;
  for(i=0;i<data.cez-2;i++) {
    for(j=0;j<NUM_SPECIES;j++)
      bufs[last+j]=vec_y[offs+j];
    last+=NUM_SPECIES;
    offs+=NUM_SPECIES*(data.cex);
  }
  
  out = mad_begin_packing(channel[1], myId+1);
  mad_pack(out, &place,sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);

   for (i=0; i<size; i++){
     p_mad_buffer_slice_parameter_t param = NULL;
    param = mad_alloc_slice_parameter();
    param->opcode = mad_op_optional_block;
    param->value  = TAUX_PERTE; /* prabibilit de perte */
    param->offset = 0; /* Pour l'instant inutile */
    param->length = 0; /* Pour l'instant inutile */
    mad_pack_ext(out,
		&bufs[i], sizeof(double),
		 mad_send_CHEAPER, mad_receive_CHEAPER,
		 param, NULL); /* donnees */
  }
  mad_end_packing(out);
  marcel_mutex_unlock(&mutexsr);
}

void send_me()
{
  p_mad_connection_t  out    = NULL;
  int i = (myId+1)%(NPEX*NPEZ);
  t_position place = ME;
  out = mad_begin_packing(channel[1], i);
  mad_pack(out, &place,sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);
  mad_end_packing(out);
}



void process_status_slist(p_tbx_slist_t status_slist, int * adresse)
{
  int compteur = 0;

  if (!tbx_slist_is_nil(status_slist)) {
    tbx_slist_ref_to_head(status_slist);
    do {
      p_mad_buffer_slice_parameter_t status = NULL;
      
      status = tbx_slist_ref_get(status_slist);
      
      if (status->opcode == mad_os_lost_block) {
	nb_lost_msgs++;
	adresse[compteur] = (int)status->base;
	compteur++;
      } else
	FAILURE("invalid or unknown status opcode");
      
    } while (tbx_slist_ref_forward(status_slist));
  }
}

static void rec_data()
{
  p_mad_connection_t  in     = NULL;
  p_tbx_slist_t       status_slist = NULL;
  int offr;
  int i,j;
  int last=0;
  double bufr[data.cez*NUM_SPECIES];
  double bufrc[(data.cex-2)*NUM_SPECIES];
  int size;
  t_position place;
  int * adresse;
  int compteur = 0;

  marcel_detach(marcel_self());
  while (quit == 0){
    
    in = mad_begin_unpacking(channel[1]);
    mad_unpack(in, &place, sizeof(t_position), mad_send_CHEAPER, mad_receive_EXPRESS);
      
    switch(place){

    case BOTTOM:
      //DISP("message du haut");
      size=(data.cex-2)*NUM_SPECIES;
      offr = NUM_SPECIES;
      for (i=0; i<size; i++)
	mad_unpack(in, &bufrc[i], sizeof(double),mad_send_CHEAPER, mad_receive_CHEAPER);
      status_slist = mad_end_unpacking_ext(in);

      if (status_slist){
	adresse = alta_malloc(size*sizeof(int));
	process_status_slist(status_slist, adresse);
	compteur = 0;
	for (i=0; i<size; i++){
	  if ((int)&bufrc[i] == adresse[compteur]){
	    compteur++;
	  }
	  else{
	    vec_y[offr + i] = bufrc[i];
	  }
	}
	alta_free(adresse);
      }
      else{
	memcpy(&vec_y[offr], bufrc,size*sizeof(double));
      }
      
      break;

    case TOP:
      //DISP("message du bas");
      offr=NEQ-NUM_SPECIES*(data.cex-1);
      size=(data.cex-2)*NUM_SPECIES;
      for (i=0; i<size; i++)
	mad_unpack(in, &bufrc[i], sizeof(double),mad_send_CHEAPER, mad_receive_CHEAPER);
      status_slist = mad_end_unpacking_ext(in);

      if (status_slist){
	adresse = alta_malloc(size*sizeof(int));
	process_status_slist(status_slist, adresse);
	compteur = 0;
	for (i=0; i<size; i++){
	  if ((int)&bufrc[i] == adresse[compteur]){
	    compteur++;
	  }
	  else{
	    vec_y[offr + i] = bufrc[i];
	  }
	}
	alta_free(adresse);
      }
      else{
	memcpy(&vec_y[offr], bufrc,size*sizeof(double));
      }
      
	
      break;
	

    case LEFT:
      //DISP("message de droite");
      size=(data.cez-2)*NUM_SPECIES;
      offr=NUM_SPECIES*(2*data.cex-1);
      for (i=0; i<size; i++)
	mad_unpack(in, &bufr[i],sizeof(double),mad_send_CHEAPER, mad_receive_CHEAPER);
      status_slist = mad_end_unpacking_ext(in);
      
      if (status_slist){
	adresse = alta_malloc(size*sizeof(int));
	process_status_slist(status_slist, adresse);
	
	last=0;
	compteur = 0;
	for(i=0;i<data.cez-2;i++) {
	  for(j=0;j<NUM_SPECIES;j++){
	    if ((int)&bufr[last+j] == adresse[compteur]){
	      compteur++;
	    }
	    else{
	      vec_y[offr+j]=bufr[last+j];
	    }
	  }
	  last+=NUM_SPECIES;
	  offr+=NUM_SPECIES*(data.cex);
	}
	
	alta_free(adresse);
      }
      else{
	last=0;
	for(i=0;i<data.cez-2;i++) {
	  for(j=0;j<NUM_SPECIES;j++)
	    vec_y[offr+j]=bufr[last+j];
	  last+=NUM_SPECIES;
	  offr+=NUM_SPECIES*(data.cex);
	}
      }
      /* 
      last=0;
      for(i=0;i<data.cez-2;i++) {
	for(j=0;j<NUM_SPECIES;j++)
	  vec_y[offr+j]=bufr[last+j];
	last+=NUM_SPECIES;
	offr+=NUM_SPECIES*(data.cex);
	}*/

      break;

    case RIGHT:
      //DISP("message de gauche");
      size=(data.cez-2)*NUM_SPECIES;
      offr=NUM_SPECIES*(data.cex);
      for (i=0; i<size; i++)
	mad_unpack(in, &bufr[i], sizeof(double),mad_send_CHEAPER, mad_receive_CHEAPER);

      status_slist = mad_end_unpacking_ext(in);
         
      if (status_slist){
	adresse = alta_malloc(size*sizeof(int));
	process_status_slist(status_slist, adresse);
	
	last=0;
	compteur = 0;
	
	for(i=0;i<data.cez-2;i++) {
	  for(j=0;j<NUM_SPECIES;j++){
	    if ((int)&bufr[last+j] == adresse[compteur]){
	      compteur++;
	    }
	    else{
	      vec_y[offr+j]=bufr[last+j];
	    }
	  }
	  last+=NUM_SPECIES;
	  offr+=NUM_SPECIES*(data.cex);
	} 
	alta_free(adresse);
      }
      else{
	last=0;
	for(i=0;i<data.cez-2;i++) {
	  for(j=0;j<NUM_SPECIES;j++)
	    vec_y[offr+j]=bufr[last+j];
	  
	  last+=NUM_SPECIES;
	  offr+=NUM_SPECIES*(data.cex);
	} 
      }
      
      /*	last=0;
	for(i=0;i<data.cez-2;i++) {
	  for(j=0;j<NUM_SPECIES;j++)
	    vec_y[offr+j]=bufr[last+j];
	  last+=NUM_SPECIES;
	  offr+=NUM_SPECIES*(data.cex);
	  }*/
	
      break;

    case ME:
      alta_printf("recu message de fin\n");
      break;

    default:
      mad_unpack(in, bufr, size*sizeof(double),mad_send_CHEAPER, mad_receive_CHEAPER);
      mad_end_unpacking(in);
      reception ++;
      break;
    }
  }
  alta_printf("thread ecoute data ferm\n");
  marcel_yield(); 
}



void* send_tt(void* a)
{
  marcel_suspend(marcel_self());
  while(!stop) {
    send_t(NULL);
    marcel_suspend(marcel_self());
  }
  alta_printf("proc %d est sorti de send t\n",myId);
  return (void*)1;
}

void* send_bb(void*a)
{
  marcel_suspend(marcel_self());
  while(!stop) {
    send_b(NULL);
    marcel_suspend(marcel_self());
  }
  alta_printf("proc %d est sorti de send b\n",myId);
  return (void*)1;
}

void* send_ll(void* a)
{
  marcel_suspend(marcel_self());
  while(!stop) {
    send_l(NULL);
    marcel_suspend(marcel_self());
  }
  alta_printf("proc %d est sorti de send l\n",myId);
  return (void*)1;
}

void* send_rr(void*a)
{
  marcel_suspend(marcel_self());
  while(!stop) {
    send_r(NULL);
    marcel_suspend(marcel_self());
  }
  alta_printf("proc %d est sorti de send r\n",myId);
  return (void*)1;
}


void convergence_detection() {
  int i;
  int flag;
	
  if(error>ERROR_THRE){
    nbIterCV=0;
    if(convergence==1){
      convergence=0;
      if(myId>0){
	if(stop_iteration==0){
	  send_c();
	}
      }else
	tab_errors[0]=error;
    }
  }else{
    nbIterCV++;
    if(nbIterCV==NBCONV){
      convergence=1;
      if(myId>0){
	if(stop_iteration==0){
	  send_c();
	}
      }else{
	tab_errors[0]=error;
	flag=1;
	for(i=0;i<nprocs && flag;++i){
	  flag&=(tab_errors[i]<ERROR_THRE);
	}
	if(flag){
	  stop_iteration=1;
	}
      }
    }
  }
  if(stop_iteration==1 && myId==0){
    stop_iteration=2;
    maxerror=error;
    for(i=1;i<nprocs;++i){
      if(maxerror<tab_errors[i])
	maxerror=tab_errors[i];
    }
    for(i=1;i<nprocs;++i){
      send_stop(i);
    }
  }
}


/****** Convergence decentralise *********/


void comptabilise_voisin(){
  int i = 0;
  /* Comme l'algorythme de detection de convergence est prvu pour
     un arbre et pas une grille, on mappe un arbre sur la grille en 
     ne prenant en compte les connexions haut/bas que sur la ligne
     x == 0 */

  nbNeighbors = 4;
  if (myId < NPEX ) nbNeighbors --;
  if (myId > (NPEX*NPEZ - NPEX - 1)) nbNeighbors --;
  if ((myId % NPEX) ==  (NPEX - 1) ) nbNeighbors --;
  if ((myId % NPEX) == 0 )nbNeighbors --;

 
  if(data.isubz!=0) { // top
    if(data.isubx != 0){
      nbNeighbors --;
    }
  }
  
  if(data.isubz!=NPEZ-1) { // bottom
    if(data.isubx != 0){
      nbNeighbors --;
    }
  }
  
  if (nbNeighbors > 0){
    neighbors = (int *) alta_malloc (nbNeighbors*sizeof(int));
    if(data.isubz!=0) { // top
      if(data.isubx == 0){
	neighbors[i] = myId - NPEX;
	i++;
      }
    }
    if(data.isubz!=NPEZ-1) { // bottom
      if(data.isubx == 0){
	neighbors[i] = myId + NPEX;
	i++;
      }
    }
    if(data.isubx!=0) { // left
      neighbors[i] = myId - 1;
      i++;
    }
    if(data.isubx!=NPEX-1) { // right
      neighbors[i] = myId + 1;
    }
  }
}

void pre_init_convergence_cd(){
  int i;
  comptabilise_voisin();
  prevIterNumS = (int *)alta_malloc(nbNeighbors*sizeof(int));
  prevIterNumC = (int *)alta_malloc(nbNeighbors*sizeof(int));
  stop_calcul  = (int *)alta_malloc(nbNeighbors*sizeof(int));

  for(i = 0; i<nbNeighbors; i++)
    {
      LDISP(" %d -> %d", myId , neighbors[i]);
    }
}

void init_convergence_cd(){
  int i;
  destinataire = -1;
  nbNeigNotLCV = nbNeighbors;
  nbIterPreLocalCV = 0;
  preLocalCV = FALSE;
  sLocalCV = FALSE;
  globalCV = FALSE;
  for(i=0; i<nbNeighbors; i++){
    prevIterNumS[i] = -1;
    prevIterNumC[i] = -1;
    stop_calcul[i]  =  0;
  }
}


void convergence_detection_cd(int iteration){
  if(!sLocalCV){
    if(error<ERROR_THRE){
      nbIterPreLocalCV++;
      if (nbIterPreLocalCV >=  THRESHOLD_SLCV ){
	sLocalCV = TRUE;
	//	LDISP("J'ai converge : %d erreur : %f", myId, error);
      }
    }
    else{
      nbIterPreLocalCV = 0;
    }
  }
  else{
    if (error>ERROR_THRE){
      sLocalCV = FALSE;
      nbIterPreLocalCV = 0;
      send_cancelPartialCV(iteration);
    }
    else{
      if(nbNeigNotLCV ==0){
	globalCV = TRUE;
	stop_iteration =1;
      }
      else{
	if(nbNeigNotLCV ==1){
	  send_PartialCV(iteration);
	}
      }
    }
  }
}

void post_convergence_cd(){
    send_GlobalCV();
}

void lib_convergence_cd(){
  alta_free(prevIterNumS);
  alta_free(prevIterNumC);
  alta_free(stop_calcul);
  alta_free(neighbors);
}



void implicit_euler_step( double* y, double* old_y, double t){
  int i,j;
  int iter;
  int result;
  int nb_col=1;
  int last=0;
    

  nbIterCV=0;
  convergence=0;

  if (T_CONV == 1) init_convergence_cd();
  
  if(myId==0) {
    for(i=0;i<nprocs;++i)
      tab_errors[i]=ERROR_THRE+1;
  }
  
  
  
  for(i=0;i<NEQ;i++)
    vec_y[i]=y[i];
  
  stop_iteration=0;
  
  for(iter=0;iter<max_iter && stop_iteration==0;iter++) {
    
    ff(vec_b,vec_y,old_y,t);

    if(iter==0) {
      nb_col=1;
      last=0;
      nb_non_zero=0;
      
      colptr[0]=0;
      for(j=0;j<NEQ;j++)
	vec_yy[j]=vec_y[j];
      
      for(i=0;i<NEQ;i++) {
	if(last!=0)
	  vec_yy[last]=vec_y[last];
	if(is_on_grid(i)) {
	  last=i;
	  //if(myId==0)
	  //	cout<<i<<endl;
	  vec_yy[i]+=hh;
	  
	  jacobian(i, tmp, vec_yy, old_y, t);
	  colptr[nb_col++]=nb_non_zero; 
	}
      }
      nb_col--;  //car il y a une colonne en plus
    }
    //		cout<<"nb col "<<myId<<" "<<nb_col<<endl;
    
    last=0;
    for(i=0;i<NEQ;i++) {
      if(is_on_grid(i)) {
	b[last]=vec_b[i];
	//if(myId==0) cout<<"b "<<i<<" "<<b[last]<<endl;
	r[last++]=0;//vec_y[i];
      }
    }
    
    result= gmres_solve(nb_col, nb_non_zero, val, colptr, rowind, b, r);
    if(result==0) {
      last=0;
      for(i=0;i<NEQ;i++){
	if(is_on_grid(i)) {
	  //cout<<"y "<<myId<<"   "<<i<<" "<<vec_y[i]<<" "<<r[last]<<" "<<vec_y[i]-r[last]<<endl;
	  vec_y[i]-=r[last++];
	  
	}
      }
      double norm=0;
      for(i=0;i<nb_col;i++)
	norm=maxloc(norm,fabs(r[i]));
      
      if(data.isubz!=0)
	if(marcel_mutex_trylock(&mutexst)==1) {
	  marcel_resume(thst);
	}
      if(data.isubz!=NPEZ-1)
	if(marcel_mutex_trylock(&mutexsb)==1) {
	  marcel_resume(thsb);
	}
      if(data.isubx!=0)
	if(marcel_mutex_trylock(&mutexsl)==1) {
	  marcel_resume(thsl);
	}
      if(data.isubx!=NPEX-1)
	if(marcel_mutex_trylock(&mutexsr)==1) {
	  marcel_resume(thsr);
	}

      error=norm; 

      if (T_CONV == 1) 
	convergence_detection_cd(iter);
      else 
	convergence_detection();
      //marcel_delay(200);
      alta_printf("%d: t=%f; i=%3d --> norm %f\n", myId, t, iter, norm);
      fflush(stdout);
    }
    else {
      alta_printf("pb dans le solveur\n");
      exit(1);
    }
  }  

  if (T_CONV == 1) post_convergence_cd(iter);
  
  if(myId==0)
  alta_printf("au temps %f il a fallu %d iterations\n",t,iter);
  
  if(iter==max_iter) {
    alta_printf("pb de convergence (iter == %d)\n", max_iter);
    exit(2);
  }
  for(i=0;i<NEQ;i++)
    y[i]=vec_y[i];  
}


void allocate_jacobian_matrix() {

  nb_non_zero=0;
  nb_non_zero=NEQ*10;
  val = alta_malloc(sizeof(double)*nb_non_zero);
  colptr  = alta_malloc(sizeof(int)*(NEQ+10));
  rowind = alta_malloc(sizeof(int)*nb_non_zero);
}


void alloc_data() {
  vec_y = alta_malloc(sizeof(double)*NEQ);
  vec_yy=alta_malloc(sizeof(double)*NEQ);
  vec_b=alta_malloc(sizeof(double)*NEQ);
  tmp=alta_malloc(sizeof(double)*NEQ);
  r=alta_malloc(sizeof(double)*NEQ);
  b=alta_malloc(sizeof(double)*NEQ);
  tmp_y=alta_malloc(sizeof(double)*NEQ);
  vec_r=alta_malloc(sizeof(double)*NEQ);
  tmp_yy=alta_malloc(sizeof(double)*NEQ);
  new_y=alta_malloc(sizeof(double)*NEQ);
  old_y=alta_malloc(sizeof(double)*NEQ);
  y=alta_malloc(sizeof(double)*NEQ);
  if(myId==0 && T_CONV == 0) tab_errors=alta_malloc(sizeof(double)*nprocs);
  if(T_CONV == 1) tab_errors=alta_malloc(sizeof(double)*nprocs);
}

void desalloc_data() {
  alta_free(vec_y);
  alta_free(vec_yy);
  alta_free(vec_b);
  alta_free(vec_r);
  alta_free(tmp);
  alta_free(r);
  alta_free(b);
  alta_free(tmp_y);
  alta_free(tmp_yy);
  alta_free(new_y);
  alta_free(old_y);
  alta_free(y);
  //  if(myId==0)
  alta_free(tab_errors);
}



static int _received_from_all_other_nodes (barrier_t *bar)
{
  int i = 0;

  barrier_lock(bar);
  while(( i < nprocs) && (bar->tab_sync[i] >= 1)) 
    i++;

  barrier_unlock(bar);
  return (i == nprocs);
}

void my_barrier(barrier_t *bar)
{
  int i;

  barrier_lock(bar);
  bar->tab_sync[myId]++;
  barrier_unlock(bar);
  
  if (T_BARR == 0){
    for (i=0 ; i < nprocs ; i++){
      if (i!=myId){
	send_bd(i);
      }
    }
  }
  else{
    if(myId > 0) { // left
      //      DISP("envoie gauche %d", myId);
      send_bdd(myId - 1, myId);
    }
    if(myId < (NPEX* NPEZ)-1) { // right
      //      DISP("envoie droite %d", myId);
      send_bdd(myId + 1, myId);
    }
  }  
  
  while (!_received_from_all_other_nodes(bar))
    marcel_yield();
  
  barrier_lock(bar);
  for (i=0 ; i < nprocs ; i++)
    bar->tab_sync[i]--;
  
  barrier_unlock(bar);
}

void my_barrier_init(barrier_t *bar)
{
  int i;
  int size = nprocs;
  
  bar->tab_sync = (int *)tmalloc(size * sizeof(int));
  marcel_mutex_init(&bar->sync_mutex, NULL);
  for (i = 0; i < size; i++)
    bar->tab_sync[i] = 0;
}




int main(int argc,char** argv) {

  int i,j,iter;
  FILE *fileu,*filev;
  char name[100];
  int jx,jz;

  p_mad_madeleine_t          madeleine    = NULL; //objet madeleine
  ntbx_process_grank_t       my_global_rank = -1;
  marcel_attr_t attr;

  common_pre_init (&argc, argv, NULL);
  common_post_init(&argc, argv, NULL);
  /* recuperation objet madeleine */
  madeleine = mad_get_madeleine();
	
	
  my_global_rank = madeleine->session->process_rank;
  myId = my_global_rank;
  nprocs =  tbx_slist_get_length(madeleine->dir->process_slist);

  /* remarque: il faut que NPEX*MXSUB=MX et idem pour Z 
   */
  /* Vecteur colonne */
  /*
    NPEX=1;
    NPEZ=4;
    MXSUB=200;
    MZSUB=50;
    MX=200;
    MZ=200;
  */
  
  /* Vecteur Ligne */
  /*
    NPEX=4;
    NPEZ=1;
    MXB=50;
    MZSUB=200;
    MX=200;
    MZ=200;
  */

  /* Grille */
  /*
    NPEX=2;
    NPEZ=2;
    MXSUB=200;
    MZSUB=200;
    MX=400;
    MZ=400;
  */

  /* exprimentation AD */
  NPEX = 2;
  NPEZ = 2;
  MXSUB = 200;
  MZSUB = 200;
  MX = MXSUB * NPEX;
  MZ = MZSUB * NPEZ;

  InitUserData();

  alloc_data();

  for(i=0;i<NEQ;i++)
    y[i]=0;

  SetInitialProfiles(y);

  for(i=0;i<NEQ;i++)
    old_y[i]=y[i];


  my_barrier_init(&barrier);
  thread_ecoute  = (marcel_t *)alta_malloc(2 * sizeof(marcel_t));
  alta_printf(" Je suis %d et il y a %d procs\n", myId, nprocs);
	
  if (nprocs != NPEX*NPEZ) {
    if (myId == 0)
      alta_printf("\n npes=%d is not equal to NPEX*NPEY=%d\n", nprocs,NPEX*NPEZ);
    return(1);
  }

  channel =alta_malloc(2 * sizeof( p_mad_channel_t));
  channel[0] =  tbx_htable_get(madeleine->channel_htable,"test");
  channel[1] =  tbx_htable_get(madeleine->channel_htable,"perte");

  if(data.isubz!=0)
    marcel_create (&thst, NULL, send_tt, NULL);
  if(data.isubz!=NPEZ-1)
    marcel_create (&thsb, NULL, send_bb, NULL);
  if(data.isubx!=0)
    marcel_create (&thsl, NULL, send_ll, NULL);
  if(data.isubx!=NPEX-1)
    marcel_create (&thsr, NULL, send_rr, NULL);

  marcel_create (&thread_ecoute[0], &attr,(marcel_func_t)rec_ordre, NULL);
  marcel_create (&thread_ecoute[1], &attr,(marcel_func_t)rec_data, NULL);

  double t = 0;
  allocate_jacobian_matrix();


  DISP("attente synchro");
  //  if (myId == 0) marcel_delay(2000);
  my_barrier(&barrier);
  DISP("Tout le monde est pret GO");

  if (T_CONV == 1) pre_init_convergence_cd();
 
  tstart=TimeStart();

  for(iter=0;iter<NOUT;iter++) {
    for(j=0;j<TWOHR/h;j++) {
      t+=h;
      my_barrier(&barrier);
      DISP("%d: t=%f; lancement calcul", myId, t);
      implicit_euler_step(y,old_y,t);
	    
      for(i=0;i<NEQ;i++) {
	old_y[i]=y[i];
      }
    }
	  
    sprintf(name,"resa1_%.3d_%.3d",iter,myId);
    fileu=fopen(name,"w");
    sprintf(name,"resa2_%.3d_%.3d",iter,myId);
    filev=fopen(name,"w");
	  
	  
	  
    for (jz=data.dsz; jz < data.dez; jz++) {
      for (jx=data.dsx; jx < data.dex; jx++) {
	fprintf(fileu,"%d %d %e\n",jx+data.bx+data.offx,jz+data.bz+data.offz,IJKth(y,1,jx+1,jz+1));
	fprintf(filev,"%d %d %e\n",jx+data.bx+data.offx,jz+data.bz+data.offz,IJKth(y,2,jx+1,jz+1));
      }
    }
	  
    fclose(fileu);
    fclose(filev);
	  
  }
	
	
  my_barrier(&barrier);
  alta_printf("temps %f\n",TimeStop(tstart));
  DISP("nombre de perte : %d", nb_lost_msgs);
  marcel_delay(1000);

  stop=1;


  //il faut etre sur que tout le monde ait mis le stop a 1 pour quitter proprement

  my_barrier(&barrier);
  //sert a debloquer les send

  if(data.isubz!=0)
    marcel_resume(thst);
  if(data.isubz!=NPEZ-1)
    marcel_resume(thsb);
  if(data.isubx!=0)
    marcel_resume(thsl);
  if(data.isubx!=NPEX-1)
    marcel_resume(thsr);

  if(data.isubz!=0)
    marcel_join (thst,NULL);
  if(data.isubz!=NPEZ-1)
    marcel_join (thsb,NULL);
  if(data.isubx!=0)
    marcel_join (thsl,NULL);
  if(data.isubx!=NPEX-1)
    marcel_join (thsr,NULL);

  DISP("tread send fermes");
  my_barrier(&barrier);

  quit = 1;

  send_me();
  marcel_delay(1000);
  marcel_delay(1000);
  send_fin();
	
  desalloc_data();

  alta_free(val);
  alta_free(colptr);
  alta_free(rowind);
  alta_free(thread_ecoute);
  alta_free(channel);
  common_exit(NULL);
  return 0;
}



